import java.awt.*;
import java.awt.geom.*;

/**
 * A user-defined area on the table.
 */
public class JuArea {

	/**
	 * You can change these constants.
	 */
	static final Color AREA_COLOR = new Color(1f, 1f, 1f, 0.08f);
	static final Color CENTER_COLOR = new Color(1f, 1f, 1, 0.08f);
	static final Stroke CENTER_STROKE = new BasicStroke(2);
	static final float CENTER_SIZE = 10;
	
	final JuTable parent;
	Shape outline = null;
	Point2D center = null;

	public JuArea(JuTable parent) {
		this.parent = parent;
	}
	
	/**
	 * Returns the center of the area. This is the point according to which documents will
	 * be oriented.
	 */
	public Point2D getCenter() {
		return center;
	}
	
	public boolean contains(Point2D p) {
		if (outline == null) return false;
		return outline.contains(p);
	}

	private void setCenter(Point2D center) {
		this.center = center;
	}
	
	public Point2D.Float getCenterFromPath(GeneralPath ink) {
		Point2D.Float p0 = getFirstPoint(ink);
		Point2D.Float p1 = getLastPoint(ink);
		// increase last point's distance by a factor 4
		float k = 4;
		p1.x += (p1.x - p0.x)*k;
		p1.y += (p1.y - p0.y)*k;
		return p1;
	}
	
	public void setCenter(GeneralPath ink) {
		setCenter(getCenterFromPath(ink));
	}
	
	public void setOutline(GeneralPath ink) {
		outline = completePath(ink);
	}
	
	public Shape getOutline() {
		return outline;
	}
	
	public void paint(Graphics2D g) {
		if (outline != null) {
			g.setColor(AREA_COLOR);
			g.setStroke(CENTER_STROKE);
			//g.fill(outline);
			g.fill(outline);
		}
		if (center != null) {
			g.setColor(CENTER_COLOR);
			g.setStroke(CENTER_STROKE);
			g.draw(new Line2D.Double(
				center.getX() - CENTER_SIZE, center.getY(),
				center.getX() + CENTER_SIZE, center.getY()));
			g.draw(new Line2D.Double(
					center.getX(), center.getY() - CENTER_SIZE,
					center.getX(), center.getY() + CENTER_SIZE));
		}
	}
	
	/**
	 * Naive method for transforming inks to closed paths.
	 */ 
	GeneralPath completePath(GeneralPath outline) {
		GeneralPath path = new GeneralPath();
		//    get the first point from the outline 
		Point2D p0 = getFirstPoint(outline);
		//    get the last point from the outline
		Point2D p1 = getLastPoint(outline);
		//    start with the extension to the first point
		Point2D p0_e = getClosestBorderPoint((float)p0.getX(), (float)p0.getY());
		path.moveTo(p0_e.getX(), p0_e.getY());
		//    add the outline
		path.append(outline, true);
		//    add the extension to the last point
		Point2D p1_e = getClosestBorderPoint((float)p1.getX(), (float)p1.getY());
		path.lineTo(p1_e.getX(), p1_e.getY());
		//    maybe add the corner
		if (p0_e.getX() != p1_e.getX() && p0_e.getY() != p1_e.getY()) {
			if (p1_e.getY() == 0 || p1_e.getY() == parent.getBounds().height)
				path.lineTo(p0_e.getX(), p1_e.getY());
			else
				path.lineTo(p1_e.getX(), p0_e.getY());
		}
		return path;
	}
	
	Point2D.Float getFirstPoint(GeneralPath path) {
		PathIterator iterator = path.getPathIterator(null);
		float[] seg = new float[4];
		iterator.currentSegment(seg);
		return new Point2D.Float(seg[0], seg[1]);
	}
	
	Point2D.Float getLastPoint(GeneralPath path) {
		return new Point2D.Float((float)path.getCurrentPoint().getX(), (float)path.getCurrentPoint().getY());
	}
	
	Point2D getClosestBorderPoint(float x, float y) {
		// -- closest border
		float x1 = parent.getBounds().width;
		float y1 = parent.getBounds().height;
		float d1 = Math.abs(x);
		float d2 = Math.abs(x-x1);
		float d3 = Math.abs(y);
		float d4 = Math.abs(y-y1);
		if (d1 <= d2 && d1 <= d3 && d1 <= d4)
			return new Point2D.Float(0, y);
		if (d2 <= d1 && d2 <= d3 && d2 <= d4)
			return new Point2D.Float(x1, y);
		if (d3 <= d1 && d3 <= d2 && d3 <= d4)
			return new Point2D.Float(x, 0);
		//if (d4 <= d1 && d4 <= d2 && d4 <= d3)
			return new Point2D.Float(x, y1);
	}
}
