import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.*;

import javax.swing.Timer;

/**
 * A paper document.
 */
public class JuDocument implements ActionListener {
	
	/**
	 * You can change the constants here.
	 */
	public static final int WIDTH = 60;
	public static final int HEIGHT = 80;
	public static final int SHADOW_OFFSET = 2;
	public static final Color BG_COLOR = new Color(0.93f, 0.93f, 0.93f, 1f); 
	public static final Color SHADOW_COLOR = new Color(0, 0, 0, 0.1f);
	public static final Color BORDER_COLOR = new Color(0, 0, 0, 0.2f);
	public static final Stroke BORDER_STROKE = new BasicStroke(1);
	public static final float ANIMATION_SPEED = 0.4f; 
	
	Timer animationTimer = new Timer(20, this);
	
	float x, y, orientation, orientation_target;
	private Shape raw_outline = null, outline = null;
	
	private AffineTransform transform = null;
	private Rectangle bounds = null;
	private final JuTable parent; // for repaint

	public JuDocument(JuTable parent) {
		this.parent = parent;
	}
	
	protected void invalidate() {
		outline = null;
		bounds = null;
		transform = null;
	}

	public Shape getUntransformedOutline() {
		if (raw_outline == null) {
			raw_outline = new Rectangle(-WIDTH/2, -HEIGHT/2, WIDTH, HEIGHT);
		}
		return raw_outline;
	}
	
	public Shape getOutline() {
		if (outline == null) {
			outline = getTransform().createTransformedShape(getUntransformedOutline());
		}
		return outline;
	}
	
	public Rectangle getBounds() {
		if (bounds == null) {
			Rectangle b = getOutline().getBounds();
			bounds = new Rectangle(b.x - 1, b.y - 1, b.width + SHADOW_OFFSET + 2, b.height + SHADOW_OFFSET + 2);
		}
		return bounds;
	}
	
	public AffineTransform getTransform() {
		if (transform == null) {
			transform = AffineTransform.getTranslateInstance(x, y);
			transform.rotate(orientation);
		}
		return transform;
	}

	
	public void paintShadow(Graphics2D g) {
		g.translate(SHADOW_OFFSET, SHADOW_OFFSET);
		g.setColor(SHADOW_COLOR);
		g.fill(getOutline());
		g.translate(-SHADOW_OFFSET, -SHADOW_OFFSET);
	}
	
	public void paint(Graphics2D g) {
		AffineTransform ot = g.getTransform();
		g.transform(getTransform());
		g.setColor(BG_COLOR);
		g.fill(getUntransformedOutline());
		g.setStroke(BORDER_STROKE);
		g.setColor(BORDER_COLOR);
		g.draw(getUntransformedOutline());
		// paint some text
		g.setColor(Color.BLACK);
		g.drawString("CV", WIDTH/60 - 10, HEIGHT/60 - 20);
		g.setColor(BORDER_COLOR);
		for (int y = -5; y < 30; y+=7) {
			g.drawLine(-18, y, 18, y);
		}
		g.setTransform(ot);
	}
	
	public void repaint() {
		parent.repaint(getBounds());
	}
	
	public void updateOrientation(boolean animate) {
		setOrientation(parent.orientationField.getOrientation(x, y), animate);
	}

	public void setLocation(float x, float y) {
		this.x = x;
		this.y = y;
		invalidate();
		updateOrientation(false);
		repaint();
	}
	
	protected void setOrientation(float angle, boolean animate) {
		if (animate && orientation != angle) {
			orientation_target = getClosestAngle(orientation, angle);
			if (!animationTimer.isRunning())
				animationTimer.start();
		} else {
			orientation = angle;
			invalidate();
			repaint();
		}
	}
	
	float getClosestAngle(float source, float dest) {
		float dest1 = dest; float d1 = Math.abs(dest1-source);
		float dest2 = dest + (float)Math.PI*2; float d2 = Math.abs(dest2-source);
		float dest3 = dest - (float)Math.PI*2; float d3 = Math.abs(dest3-source);
		if (d1 <= d2 && d1 <= d3) return dest1;
		if (d2 <= d1 && d2 <= d3) return dest2;
		return dest3;
	}
	
	public void drag(Point2D p0, Point2D p1) {
		repaint();
		x += p1.getX() - p0.getX();
		y += p1.getY() - p0.getY();
		updateOrientation(false);
		invalidate();
		repaint();
	}

	// For the orientation animation
	public void actionPerformed(ActionEvent arg0) {
		// stop animation?
		if (Math.abs(orientation - orientation_target) < 0.01) {
			repaint();
			orientation = orientation_target;
			if (orientation > Math.PI*2) orientation -= Math.PI*2;
			if (orientation < 0) orientation += Math.PI*2;
			invalidate();
			repaint();
			animationTimer.stop();
			return;
		}
		// animate
		repaint();
		orientation += (orientation_target - orientation)*ANIMATION_SPEED;
		invalidate();
		repaint();
	}

}
