import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;

/**
 * Class for managing document orientations.
 * 
 * Contains methods for automatically computing document orientations, and methods
 * for creating, modifying and displaying a vector field. 
 */
public class JuOrientationField {
	
	/**
	 * You can change the constants here.
	 */
	static final int FIELD_RESOLUTION_X = 100;
	static final int FIELD_RESOLUTION_Y = 70;
	static final float COLOR_SUBSAMPLING = 2f; // for display
	static final float VECTOR_SUBSAMPLING = 0.25f; // for display
	static final int SMOOTHING = 8; // the radius of the averaging window
	public static final Color VECTOR_COLOR = new Color(1f, 1f, 1f, 0.1f);
	public static final Stroke VECTOR_STROKE = new BasicStroke(2); 
	
	JuTable parent;
	Rectangle tableBounds;
	int orientationMethod = 0;
	boolean fieldEnabled = false;
	boolean smoothingEnabled = false;
	boolean showColors = false;
	boolean showVectors = false;
	
	// The vector field
	final Point2D.Float[][] field = new Point2D.Float[FIELD_RESOLUTION_X][FIELD_RESOLUTION_Y];
	{for (int i=0; i<FIELD_RESOLUTION_X; i++)
		for (int j=0; j<FIELD_RESOLUTION_Y; j++)
			field[i][j] = new Point2D.Float();}
	// A temporary vector field used by the smoothing algorithm
	final Point2D.Float[][] tmp = new Point2D.Float[FIELD_RESOLUTION_X][FIELD_RESOLUTION_Y];
	{for (int i=0; i<FIELD_RESOLUTION_X; i++)
		for (int j=0; j<FIELD_RESOLUTION_Y; j++)
			tmp[i][j] = new Point2D.Float();}
	// Image into which the vector field is pre-rendered
	BufferedImage bitmap;
	
	public JuOrientationField(JuTable parent) {
		this.parent = parent;
		tableBounds = parent.getTableBounds();
		bitmap = new BufferedImage(tableBounds.width, tableBounds.height, BufferedImage.TYPE_INT_ARGB);
	}
	
	public void invalidateField() {
		if (!fieldEnabled) return;
		// compute the field
		for (int i=0; i<FIELD_RESOLUTION_X; i++)
			for (int j=0; j<FIELD_RESOLUTION_Y; j++) {
				field[i][j].x = getOrientationFromFormula(i2x(i), j2y(j));
				field[i][j].y = 10;
				toRect(field[i][j]);
		}	
		// maybe smooth the field
		if (smoothingEnabled)
			smoothField(SMOOTHING);
		// buffer the visualization
		invalidateVisualization();
		return;
	}
	
	public void invalidateVisualization() {
		dopaint((Graphics2D)bitmap.getGraphics());
	}
	
	/**
	 * Smoothes the vector field
	 */
	public void smoothField(int amount) {
		for (int i=0; i<FIELD_RESOLUTION_X; i++) {
			for (int j=0; j<FIELD_RESOLUTION_Y; j++) {
				// compute the average
				int sumx=0, sumy=0, samples=0;
				for (int i0 = i-amount; i0 <= i+amount; i0++) {
					for (int j0 = j-amount; j0 <= j+amount; j0++) {
						if (i0 >= 0 && j0 >=0 && i0 < FIELD_RESOLUTION_X && j0 < FIELD_RESOLUTION_Y) {
							sumx += field[i0][j0].x;
							sumy += field[i0][j0].y;
							samples++;
						}
					}
				}
				// record the average in the temporary vector field
				if (samples > 0) {
					tmp[i][j].x = sumx/(float)samples;
					tmp[i][j].y = sumy/(float)samples;
				}
			}
		}
		// copy the temporary vector field in the actual vector field
		for (int i=0; i<FIELD_RESOLUTION_X; i++)
			for (int j=0; j<FIELD_RESOLUTION_Y; j++) {
				field[i][j].x = tmp[i][j].x;
				field[i][j].y = tmp[i][j].y;
			}
	}
	
	
	// Conversion from flow array index to screen coordinates and back
	float i2x(float i) {return tableBounds.x + i*tableBounds.width/(float)FIELD_RESOLUTION_X;}
	float j2y(float j) {return tableBounds.y + j*tableBounds.height/(float)FIELD_RESOLUTION_Y;}
	float x2i(float x) {return (x-tableBounds.x)*FIELD_RESOLUTION_X/(float)tableBounds.width;}
	float y2j(float y) {return (y-tableBounds.y)*FIELD_RESOLUTION_Y/(float)tableBounds.height;}
	
	// Binilear interpolation of the vector field
	public void interpolateVectorAt(float x, float y, Point2D.Float output) {
		float i = x2i(x);
		float j = y2j(y);
		// compute indices for bilinear interpolation
		int i0 = (int)i; int i1 = i0+1; float it = i-i0;
		if (i0 < 0) i0 = 0; if (i0 >= FIELD_RESOLUTION_X) i0 = FIELD_RESOLUTION_X - 1;
		if (i1 < 0) i1 = 0; if (i1 >= FIELD_RESOLUTION_X) i1 = FIELD_RESOLUTION_X - 1;
		int j0 = (int)j; int j1 = j0+1; float jt = j-j0;
		if (j0 < 0) j0 = 0; if (j0 >= FIELD_RESOLUTION_Y) j0 = FIELD_RESOLUTION_Y - 1;
		if (j1 < 0) j1 = 0; if (j1 >= FIELD_RESOLUTION_Y) j1 = FIELD_RESOLUTION_Y - 1;
		// bilinear interpolation of x
		float ix0 = (1-jt) * field[i0][j0].x + jt * field[i0][j1].x;
		float ix1 = (1-jt) * field[i1][j0].x + jt * field[i1][j1].x;
		float ix = (1-it) * ix0 + it * ix1;
		// bilinear interpolation of y
		float iy0 = (1-jt) * field[i0][j0].y + jt * field[i0][j1].y;
		float iy1 = (1-jt) * field[i1][j0].y + jt * field[i1][j1].y;
		float iy = (1-it) * iy0 + it * iy1;
		//
		output.x = ix;
		output.y = iy;
	}
	
	// ---- Orientation methods
	
	public void setOrientationMethod(int method) {
		orientationMethod = method;
	}
	public int getOrientationMethod() {
		return orientationMethod;
	}
	
	public float getOrientation(float x, float y) {
		if (fieldEnabled)
			return getOrientationFromField(x, y);
		else
			return getOrientationFromFormula(x, y);
	}
	
	public float getOrientationFromField(float x, float y) {
		Point2D.Float p = new Point2D.Float();
		interpolateVectorAt(x, y, p);
		return (float)angle(new Point(0, 0), p);
	}
	
	public float getOrientationFromFormula(float x, float y) {
		switch (getOrientationMethod()) {
		case 0:
			return orientStraight(x, y);
		case 1:
			return orientCircle(x, y);
		case 2:
			return orientEllipse(x, y);
		case 3:
			return orientBorder(x, y);
		case 4:
			return orientAreas(x, y);
		}
		
		return 0;
	}
	
	// -- straight
	public float orientStraight(float x, float y) {
		return 0;
	}
	
	protected float orientTowards(float x, float y, Point2D to) {
		float angle = (float)angle(new Point2D.Float(x, y), to);
		return angle - (float)Math.PI / 2;
	}

	// -- circle
	public float orientCircle(float x, float y) {
		float midx = tableBounds.x + tableBounds.width / 2;
		float midy = tableBounds.y + tableBounds.height / 2;
		float angle = orientTowards(x, y, new Point2D.Float(midx, midy));
		return angle + (float)Math.PI;
	}
	
	// -- ellipse
	public float orientEllipse(float x, float y) {
		float midx = tableBounds.x + tableBounds.width / 2;
		float midy = tableBounds.y + tableBounds.height / 2;
		float y2 = midy + (y - midy) * (tableBounds.width / (float)tableBounds.height);
		float angle = (float)angle(new Point2D.Float(x, y2), new Point2D.Float(midx, midy));
		return angle + (float)Math.PI / 2;
	}

	// -- closest border
	public float orientBorder(float x, float y) {
		float x0 = tableBounds.x;
		float y0 = tableBounds.y;
		float x1 = tableBounds.x + tableBounds.width;
		float y1 = tableBounds.y + tableBounds.height;
		float d1 = Math.abs(x-x0);
		float d2 = Math.abs(x-x1);
		float d3 = Math.abs(y-y0);
		float d4 = Math.abs(y-y1);
		float angle = 0;
		if (d1 <= d2 && d1 <= d3 && d1 <= d4)
			angle = (float)0;
		if (d2 <= d1 && d2 <= d3 && d2 <= d4)
			angle = (float)Math.PI;
		if (d3 <= d1 && d3 <= d2 && d3 <= d4)
			angle = (float)Math.PI/2;
		if (d4 <= d1 && d4 <= d2 && d4 <= d3)
			angle = (float)-Math.PI/2;
		return angle + (float)Math.PI / 2;
	}
	
	// -- orient by taking user-defined areas into account
	public float orientAreas(float x, float y) {
		JuArea area = parent.pickArea(x, y);
		if (area != null) {
			Point2D loc = area.getCenter();
			if (loc != null)
				return orientTowards(x, y, loc);
			else
				return 0;
		}
		return orientBorder(x, y); // default
	}
	
	// angle between two points
    static double angle(Point2D a, Point2D b) {
        double dx = b.getX() - a.getX();
        double dy = b.getY() - a.getY();
        double angle = 0.0d;
        if (dx == 0.0) {
            if(dy == 0.0)     angle = 0.0;
            else if(dy > 0.0) angle = Math.PI / 2.0;
            else              angle = (Math.PI * 3.0) / 2.0;
        }
        else if(dy == 0.0) {
            if(dx > 0.0)      angle = 0.0;
            else              angle = Math.PI;
        }
        else {
            if(dx < 0.0)      angle = Math.atan(dy/dx) + Math.PI;
            else if(dy < 0.0) angle = Math.atan(dy/dx) + (2*Math.PI);
            else              angle = Math.atan(dy/dx);
        }
        return angle;
    }
    
    // rectangular to polar conversion
    static void toPolar(Point2D.Float p) {
    	float angle = (float)angle(new Point(0, 0), p);
    	float r = (float)Math.sqrt(p.x*p.x + p.y*p.y);
    	p.x = angle;
    	p.y = r;
    }
    
    // polar to rectangular conversion
    static void toRect(Point2D.Float p) {
    	float x = p.y*(float)Math.cos(p.x);
    	float y = p.y*(float)Math.sin(p.x);
    	p.x = x;
    	p.y = y;
    }
    
    public void paint(Graphics2D g) {
    	if (!fieldEnabled) return;
    	//dopaint(g);
    	g.drawImage(bitmap, parent.MARGIN, parent.MARGIN, null);
    }
    
    // -- Field visualization method
    public void dopaint(Graphics2D g) {
    	if (!fieldEnabled) return;
    		
    	float width = tableBounds.width;
    	float height = tableBounds.height;
    	Point2D.Float v = new Point2D.Float();
    	float stepx, stepy;
    	
		g.setColor(JuTable.TABLE_BG_COLOR);
		g.fillRect(0, 0, (int)width+1, (int)height+1);
    	
    	if (showColors) {
	    	stepx = width/(float)FIELD_RESOLUTION_X/COLOR_SUBSAMPLING;
	    	stepy = height/(float)FIELD_RESOLUTION_Y/COLOR_SUBSAMPLING;   	
	    	Point2D.Float o = new Point2D.Float(0,0);
			for (float x=0; x<=width; x+=stepx) {
				for (float y=0; y<height; y+=stepy) {
					interpolateVectorAt(x + tableBounds.x, y + tableBounds.y, v);
					float angle = (float)angle(o, v);
					float hue = angle/(float)(2*Math.PI);
					g.setColor(new Color(Color.HSBtoRGB(hue, 0.5f, 0.5f)));
					g.fill(new Rectangle2D.Float(x-stepx/2f, y-stepy/2f, stepx+1, stepy+1));
				}
			}
    	}

    	if (showVectors) {
	    	stepx = width/(float)FIELD_RESOLUTION_X/VECTOR_SUBSAMPLING;
	    	stepy = height/(float)FIELD_RESOLUTION_Y/VECTOR_SUBSAMPLING;   	
	    	g.setStroke(VECTOR_STROKE);
	    	g.setColor(VECTOR_COLOR);
			for (float x=0; x<=width; x+=stepx) {
				for (float y=0; y<height; y+=stepy) {
					interpolateVectorAt(x + tableBounds.x, y + tableBounds.y, v);
					g.draw(new Line2D.Float(x, y, x+v.x, y+v.y));
				}
			}
    	}
		  	
    /*	for (float a=0.4f; a>0.1f; a-=0.1f) {
    	float x = width*a;
    	float y = height*a;
    	for (int i=0; i<500; i++) {
    		getVectorAt(x, y, v);
    		g.draw(new Line2D.Float(x, y, x+v.x, y+v.y));
    		x = x+v.x;
    		y = y+v.y;
    	}
    	}*/
    	
    /*	for (float a=0f; a<Math.PI*2; a+=Math.PI/20) {
        	float x = width/2 + 10*(float)Math.cos(a);
        	float y = height/2 + 10*(float)Math.sin(a);
        	for (int i=0; i<30; i++) {
        		getVectorAt(x, y, v);
        		g.draw(new Line2D.Float(x, y, x-v.y, y+v.x));
        		x = x-v.y;
        		y = y+v.x;
        	}
        }*/
    	
    }
}
