import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.awt.image.MemoryImageSource;
import java.util.*;

/**
 * The JComponent that does everything.
 * 
 * Manages documents (JuDocument), user-defined areas (JuArea),
 * user avatars (JuGuys), and an orientation field (JuOrientationField).  
 */
public class JuTable extends JComponent implements MouseListener, MouseMotionListener, KeyListener {
	
	private static final long serialVersionUID = 1L;

	/**
	 * You can change the constants here.
	 */
	public static final int MARGIN = 120; 	// free space around the table
	public static final int BORDER_WIDTH = 5; // table border
	public static final int RAINBOW_STEPS = 50;
	public static final int RAINBOW_RADIUS = 50;
	public static final int RAINBOW_WIDTH = 5;
	public static final Color BG_COLOR = new Color(0.9f, 0.9f, 0.9f); 
	public static final Color TABLE_BG_COLOR = new Color(0.3f, 0.3f, 0.3f); 
	public static final Color INK_COLOR = new Color(0.3f, 0.3f, 0.95f, 0.5f); 
	public static final Stroke INK_STROKE = new BasicStroke(2f);
	
	// Define an invisible cursor
	final static int[] pixels = new int[16 * 16];
	final static Image image = Toolkit.getDefaultToolkit().createImage(
	        new MemoryImageSource(16, 16, pixels, 0, 16));
	final static Cursor transparentCursor =
	        Toolkit.getDefaultToolkit().createCustomCursor
	             (image, new Point(0, 0), "invisibleCursor");
	final static Cursor defaultCursor = new Cursor(Cursor.HAND_CURSOR);

	private final Vector documents = new Vector(); 
	private final Vector areas = new Vector(); 
	private final Vector guys = new Vector(); 
	final JuOrientationField orientationField;

	int width, height; // the component size
	boolean showColorBorder = false;
	boolean showRainbow = false;
	JuDocument pickedDoc = null;
	JuGuy pickedGuy = null;
	Point lastdrag = null;
	Point2D.Float laserpointer = new Point2D.Float();
	boolean dragging = false;
	GeneralPath ink = null; // ink feedback
	
	public JuTable(int tablewidth, int tableheight) {
		this.width = tablewidth + MARGIN;
		this.height = tableheight + MARGIN;
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(this);
		orientationField = new JuOrientationField(this);
		orientationField.invalidateField();
		setCursor(transparentCursor);
	}
	
	public Rectangle getBounds() {
		return new Rectangle(0, 0, width, height);
	}
	
	public Rectangle getTableBounds() {
		return new Rectangle(MARGIN, MARGIN, width-MARGIN*2, height-MARGIN*2);
	}
	
	public void repaintLaser() {
		repaint((int)laserpointer.x - 10, (int)laserpointer.y - 10, 21, 21);
		if (showRainbow)
			repaint((int)laserpointer.x - RAINBOW_RADIUS - RAINBOW_WIDTH - 5, (int)laserpointer.y - RAINBOW_RADIUS - RAINBOW_WIDTH - 5, 2*(RAINBOW_RADIUS+RAINBOW_WIDTH) + 10, 2*(RAINBOW_RADIUS+RAINBOW_WIDTH) + 10);
	}
	
	public void paint(Graphics g_) {
		Graphics2D g = (Graphics2D)g_;
		
		// -- Background
		g.setColor(BG_COLOR);
		g.fill(getBounds());
		
		// -- Table's shadow
		g.setColor(JuDocument.SHADOW_COLOR);
		g.fillRect(getTableBounds().x + 20, getTableBounds().y + 20, getTableBounds().width, getTableBounds().height);

		// -- Avatars (graphics displayed under the table)
		for (int i=0; i<guys.size(); i++) {
			JuGuy guy = (JuGuy)guys.elementAt(i);
			guy.paint_bg(g);
		}
		
		// -- Table
		g.setColor(TABLE_BG_COLOR);
		g.fill(getTableBounds());

		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		Shape oldclip = g.getClip();
		g.clip(getTableBounds());
		
		// -- Vector field
		orientationField.paint(g);

		// -- Areas
		JuArea current = getLastAddedArea();
		if (current != null && current.center != null) current = null;
		for (int i=0; i<areas.size(); i++) {
			JuArea area = (JuArea)areas.elementAt(i);
			if (area == current || !orientationField.showColors)
				area.paint(g);
		}
		
		// -- Ink
		if (ink != null) {
			g.setColor(INK_COLOR);
			g.setStroke(INK_STROKE);
			g.draw(ink);
		}

		// -- Documents
		for (int i=0; i<documents.size(); i++) {
			JuDocument doc = (JuDocument)documents.elementAt(i);
			doc.paintShadow(g);
			doc.paint(g);
		}
		
		// -- Rainbow
		if (laserpointer != null && showRainbow && pickedDoc != null) {
			g.setStroke(new BasicStroke(RAINBOW_WIDTH));
			for (int i=0; i<RAINBOW_STEPS; i++) {
				float a0 = (float)360*i/RAINBOW_STEPS;
				float a1 = (float)360*(i+1)/RAINBOW_STEPS;
				Color c = new Color(Color.HSBtoRGB(0.75f-(a0+a1)/2/360, 0.5f, 0.5f));
				g.setColor(c);//new Color(c.getRed(), c.getGreen(), c.getBlue(), 100));
				g.draw(new Arc2D.Float(laserpointer.x - RAINBOW_RADIUS, laserpointer.y - RAINBOW_RADIUS,
						RAINBOW_RADIUS*2, RAINBOW_RADIUS*2,
						a0, a1-a0, Arc2D.OPEN));
			}
			if (pickedDoc != null) {
				g.setStroke(new BasicStroke(1));
				g.setColor(new Color(0, 0, 0, 0.5f));
				float angle = pickedDoc.orientation + (float)Math.PI/2;
				GeneralPath highlight = new GeneralPath();
				float w = 0.08f;
				highlight.moveTo(laserpointer.x + Math.cos(angle-w) * (RAINBOW_RADIUS-RAINBOW_WIDTH/2-3), laserpointer.y + Math.sin(angle-w) * (RAINBOW_RADIUS-RAINBOW_WIDTH/2-3));
				highlight.lineTo(laserpointer.x + Math.cos(angle-w) * (RAINBOW_RADIUS+RAINBOW_WIDTH/2+3), laserpointer.y + Math.sin(angle-w) * (RAINBOW_RADIUS+RAINBOW_WIDTH/2+3));
				highlight.lineTo(laserpointer.x + Math.cos(angle+w) * (RAINBOW_RADIUS+RAINBOW_WIDTH/2+3), laserpointer.y + Math.sin(angle+w) * (RAINBOW_RADIUS+RAINBOW_WIDTH/2+3));
				highlight.lineTo(laserpointer.x + Math.cos(angle+w) * (RAINBOW_RADIUS-RAINBOW_WIDTH/2-3), laserpointer.y + Math.sin(angle+w) * (RAINBOW_RADIUS-RAINBOW_WIDTH/2-3));
				highlight.lineTo(laserpointer.x + Math.cos(angle-w) * (RAINBOW_RADIUS-RAINBOW_WIDTH/2-3), laserpointer.y + Math.sin(angle-w) * (RAINBOW_RADIUS-RAINBOW_WIDTH/2-3));
				g.draw(highlight);
			}
		}
		
		// -- Table border
		g.setColor(Color.GRAY);
		if (showColorBorder) g.setColor(new Color(Color.HSBtoRGB(3/4f, 1f, 0.9f)));
		g.fillRect(MARGIN + width - MARGIN*2 - BORDER_WIDTH, MARGIN, BORDER_WIDTH, height - MARGIN*2);
		if (showColorBorder) g.setColor(new Color(Color.HSBtoRGB(2/4f, 1f, 0.9f)));
		g.fillRect(MARGIN, MARGIN, width - MARGIN*2, BORDER_WIDTH);
		if (showColorBorder) g.setColor(new Color(Color.HSBtoRGB(1/4f, 1f, 0.9f)));
		g.fillRect(MARGIN, MARGIN, BORDER_WIDTH, height - MARGIN*2);
		if (showColorBorder) g.setColor(new Color(Color.HSBtoRGB(0/4f, 1f, 0.9f)));
		g.fillRect(MARGIN, MARGIN + height - MARGIN*2 - BORDER_WIDTH, width - MARGIN*2, BORDER_WIDTH);
		g.setColor(Color.black);
		g.setStroke(new BasicStroke(2));
		g.drawRect(MARGIN + BORDER_WIDTH, MARGIN + BORDER_WIDTH, width - MARGIN*2 - BORDER_WIDTH*2, height - MARGIN*2 - BORDER_WIDTH*2);
		
		g.setClip(oldclip);

		// -- Avatars (graphics displayed above the table)
		for (int i=0; i<guys.size(); i++) {
			JuGuy guy = (JuGuy)guys.elementAt(i);
			guy.paint_fg(g);
		}

		// -- "New avatar" icon
		g.drawImage(JuGuy.guy_icon, 5, height - 5 - JuGuy.icon_height, null);
		
		// -- Laser beam
		if (laserpointer != null && getTableBounds().contains(laserpointer)) {
			if (dragging) {
				g.setStroke(INK_STROKE);
				g.setColor(new Color(0.3f, 0.3f, 0.95f, 0.5f));
				g.fill(new Ellipse2D.Float(laserpointer.x - 8, laserpointer.y - 8, 17, 17));
			}
			g.setColor(new Color(1f, 0.2f, 0.2f, 0.6f));
			g.fill(new Ellipse2D.Float(laserpointer.x - 4, laserpointer.y - 4, 9, 9));
			g.setColor(new Color(1f, 0.5f, 0.5f, 0.8f));
			g.fill(new Ellipse2D.Float(laserpointer.x - 2, laserpointer.y - 2, 5, 5));
			g.setColor(new Color(1f, 0.8f, 0.8f, 1f));
			g.fill(new Ellipse2D.Float(laserpointer.x - 1, laserpointer.y - 1, 3, 3));
		}
		
		
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
	}
	
	public void addDocument(float x, float y) {
		JuDocument doc = new JuDocument(this);
		doc.setLocation(x, y);
		documents.add(doc);
		doc.repaint();
	}
	
	public JuGuy addGuy(float x, float y) {
		JuGuy guy = new JuGuy(this);
		guy.setLocation(x, y);
		guys.add(guy);
		guy.repaint();
		return guy;
	}
	
	JuDocument pickDocument(float x, float y) {
		for (int i=documents.size()-1; i>=0; i--) {
			JuDocument doc = (JuDocument)documents.elementAt(i);
			if (doc.getOutline().contains(x, y))
				return doc;
		}
		return null;
	}
	
	JuArea pickArea(float x, float y) {
		for (int i=areas.size()-1; i>=0; i--) {
			JuArea area = (JuArea)areas.elementAt(i);
			if (area.getOutline().contains(x, y))
				return area;
		}
		return null;
	}
	
	JuGuy pickGuy(float x, float y) {
		for (int i=guys.size()-1; i>=0; i--) {
			JuGuy guy = (JuGuy)guys.elementAt(i);
			if (guy.getBounds().contains(x, y))
				return guy;
		}
		return null;
	}
	
	boolean pickGuyIcon(float x, float y) {
		return (x >= 5 && x <= 5 + JuGuy.icon_width && y >= height - 5 - JuGuy.icon_height && y <= height - 5);
	}
	
	public Dimension getPreferredSize() {
		return new Dimension(width, height);
	}
	
	JuArea getLastAddedArea() {
		if (areas.size() > 0)
			return (JuArea)areas.lastElement();
		return null;
	}
	
	void updateDocumentsOrientation() {
		for (int i=documents.size()-1; i>=0; i--) {
			JuDocument doc = (JuDocument)documents.elementAt(i);
			doc.updateOrientation(true);
		}
	}
	
	public void setOrientationMethod(int method) {
		orientationField.setOrientationMethod(method); 
		orientationField.invalidateField();
		updateDocumentsOrientation();
		repaint();
	}
	
	public void setVectorFieldEnabled(boolean enabled) {
		orientationField.fieldEnabled = enabled;
		orientationField.invalidateField();
		updateDocumentsOrientation();
		repaint();
	}
	
	public void setSmoothFieldEnabled(boolean enabled) {
		orientationField.smoothingEnabled = enabled;
		orientationField.invalidateField();
		updateDocumentsOrientation();
		repaint();
	}
	
	public void showColors(boolean show) {
		orientationField.showColors = show;
		orientationField.invalidateVisualization();
		repaint();
	}
	
	public void showVectors(boolean show) {
		orientationField.showVectors = show;
		orientationField.invalidateVisualization();
		repaint();
	}

	public void mouseClicked(MouseEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	public void mousePressed(MouseEvent arg0) {
		dragging = true;
		laserpointer.setLocation(arg0.getX(), arg0.getY());
		repaintLaser();
		
		if (pickGuyIcon(arg0.getX(), arg0.getY())) {
			pickedGuy = addGuy(arg0.getX(), arg0.getY());
			pickedGuy.visible = false;
		} else if (getTableBounds().contains(arg0.getPoint())) {
			pickedDoc = pickDocument(arg0.getX(), arg0.getY());
		} else {
			pickedGuy = pickGuy(arg0.getX(), arg0.getY());
		}
		if (pickedDoc != null) {
			documents.remove(pickedDoc);
			documents.add(pickedDoc);
			pickedDoc.repaint();
			lastdrag = arg0.getPoint();
		} else if (pickedGuy != null) {
			pickedGuy.repaint();
			lastdrag = arg0.getPoint();
		} else {
			ink = new GeneralPath();
			ink.moveTo(arg0.getX(), arg0.getY());
		}
	}

	public void mouseDragged(MouseEvent arg0) {
		repaintLaser();
		laserpointer.setLocation(arg0.getPoint());
		repaintLaser();
		if (pickedDoc != null) {
			// -- Drag document
			pickedDoc.drag(lastdrag, arg0.getPoint());
			lastdrag = arg0.getPoint();
		}
		if (pickedGuy != null) {
			// -- Drag the guy
			if (SwingUtilities.isLeftMouseButton(arg0))
				pickedGuy.drag(lastdrag, arg0.getPoint());
			else
				pickedGuy.rotate(lastdrag, arg0.getPoint());
			pickedGuy.visible = !pickGuyIcon(arg0.getX(), arg0.getY());
			lastdrag = arg0.getPoint();
		}
		if (ink != null) {
			// -- Ink
			ink.lineTo(arg0.getX(), arg0.getY());
			repaint(ink.getBounds());
		}
		if (getTableBounds().contains(arg0.getPoint()) && getCursor() != transparentCursor)
			setCursor(transparentCursor);
		if (!getTableBounds().contains(arg0.getPoint()) && getCursor() != defaultCursor)
			setCursor(defaultCursor);
	}
	
	public void mouseReleased(MouseEvent arg0) {
		repaintLaser();
		if (ink != null) {
			JuArea area = getLastAddedArea();
			if (area == null || area.center != null) {
				// -- Add a new area and define its border
				area = new JuArea(this);
				areas.add(area);
				area.setOutline(ink);
			} else {
				// -- Define the center of the previous area
				area.setCenter(ink);
				orientationField.invalidateField();
				updateDocumentsOrientation();
			}
			repaint();
		}
		
		if (pickedGuy != null && !pickedGuy.visible) {
			pickedGuy.repaint();
			guys.remove(pickedGuy);
		}			
		
		dragging = false;
		pickedDoc = null;
		pickedGuy = null;
		lastdrag = null;
		ink = null;
	}

	public void mouseEntered(MouseEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	public void mouseExited(MouseEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	public void mouseMoved(MouseEvent arg0) {
		repaintLaser();
		laserpointer.setLocation(arg0.getPoint());
		repaintLaser();
		if (getTableBounds().contains(arg0.getPoint()) && getCursor() != transparentCursor)
			setCursor(transparentCursor);
		if (!getTableBounds().contains(arg0.getPoint()) && getCursor() != defaultCursor)
			setCursor(defaultCursor);
	}

	public void keyTyped(KeyEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	public void keyPressed(KeyEvent arg0) {
		if (arg0.getKeyCode() >= KeyEvent.VK_0 && arg0.getKeyCode() <= KeyEvent.VK_9) {
			setOrientationMethod(arg0.getKeyCode() - KeyEvent.VK_0); 
		}
	}

	public void keyReleased(KeyEvent arg0) {
		// TODO Auto-generated method stub
		
	}
}
